-- ----------------------------------------------------------------------
-- slony1_funcs.v93.sql
--
--    Version 9.3 specific part of the replication support functions.
--
--	Author: RBT.
--
--
-- ----------------------------------------------------------------------
-- ----------------------------------------------------------------------
-- FUNCTION ShouldSlonyVacuumTable (nspname, tabname)
--
--	Returns 't' if the table needs to be vacuumed by Slony-I
--      Returns 'f' if autovac handles the table, so Slony-I should not
--                  or if the table is not needful altogether
-- ----------------------------------------------------------------------
create or replace function @NAMESPACE@.ShouldSlonyVacuumTable (i_nspname name, i_tblname name) returns boolean as
$$
declare
	c_table oid;
	c_namespace oid;
	c_enabled boolean;
	v_dummy int4;
begin
	if not exists (select 1 from pg_catalog.pg_class c, pg_catalog.pg_namespace n
					where c.relname = i_tblname and n.nspname = i_nspname and c.relnamespace = n.oid) then
		return 'f'::boolean;   -- if table does not exist, then don't vacuum
	end if;
	select 1 into v_dummy from "pg_catalog".pg_settings where name = 'autovacuum' and setting = 'on';
	if not found then
		return 't'::boolean;       -- If autovac is turned off, then we gotta vacuum
	end if;

	select into c_namespace oid from "pg_catalog".pg_namespace where nspname = i_nspname;
	if not found then
		raise exception 'Slony-I: namespace % does not exist', i_nspname;
	end if;
	select into c_table oid from "pg_catalog".pg_class where relname = i_tblname and relnamespace = c_namespace;
	if not found then
		raise warning 'Slony-I: table % does not exist in namespace %/%', i_tblname, c_namespace, i_nspname;
		return 'f'::boolean;
	end if;

	-- So, the table is legit; try to look it up for autovacuum policy
	if exists (select 1 from pg_class where 'autovacuum_enabled=off' = any (reloptions) and oid = c_table) then
		return 't'::boolean;   -- Autovac is turned on, but this table is disabled
	end if;

	return 'f'::boolean;

end;$$ language plpgsql;

comment on function @NAMESPACE@.ShouldSlonyVacuumTable (i_nspname name, i_tblname name) is
'returns false if autovacuum handles vacuuming of the table, or if the table does not exist; returns true if Slony-I should manage it';

create or replace function @NAMESPACE@.TruncateOnlyTable (name) returns void as
$$
begin
	execute 'truncate only '|| @NAMESPACE@.slon_quote_input($1);
end;
$$
LANGUAGE plpgsql;

comment on function @NAMESPACE@.TruncateOnlyTable(name) is
'Calls TRUNCATE ONLY, syntax supported in version >= 8.4';

create or replace function @NAMESPACE@.alterTableAddTruncateTrigger (i_fqtable text, i_tabid integer) returns integer as $$
begin
		execute 'create trigger "_@CLUSTERNAME@_truncatetrigger" ' ||
				' before truncate on ' || i_fqtable || ' for each statement execute procedure ' ||
				'@NAMESPACE@.log_truncate(' || i_tabid || ');';
		execute 'create trigger "_@CLUSTERNAME@_truncatedeny" ' ||
				' before truncate on ' || i_fqtable || ' for each statement execute procedure ' ||
				'@NAMESPACE@.deny_truncate();';
		return 1;
end
$$ language plpgsql;

comment on function @NAMESPACE@.alterTableAddTruncateTrigger (i_fqtable text, i_tabid integer) is
'function to add TRUNCATE TRIGGER';

create or replace function @NAMESPACE@.alterTableDropTruncateTrigger (i_fqtable text, i_tabid integer) returns integer as $$
begin
		execute 'drop trigger "_@CLUSTERNAME@_truncatetrigger" ' ||
				' on ' || i_fqtable || ';';
		execute 'drop trigger "_@CLUSTERNAME@_truncatedeny" ' ||
				' on ' || i_fqtable || ';';
		return 1;
end
$$ language plpgsql;

comment on function @NAMESPACE@.alterTableDropTruncateTrigger (i_fqtable text, i_tabid integer) is
'function to drop TRUNCATE TRIGGER';

create or replace function @NAMESPACE@.alterTableConfigureTruncateTrigger(i_fqname text, i_log_stat text, i_deny_stat text) returns integer as $$
begin
		execute 'alter table ' || i_fqname || ' ' || i_log_stat ||
				' trigger "_@CLUSTERNAME@_truncatetrigger";';
		execute 'alter table ' || i_fqname || ' ' || i_deny_stat ||
				' trigger "_@CLUSTERNAME@_truncatedeny";';
		return 1;
end $$ language plpgsql;

comment on function @NAMESPACE@.alterTableConfigureTruncateTrigger(i_fqname text, i_log_stat text, i_deny_stat text) is
'Configure the truncate triggers according to origin status.';

create or replace function @NAMESPACE@.upgradeSchemaAddTruncateTriggers () returns integer as $$
begin

		--- Add truncate triggers
		begin
		perform @NAMESPACE@.alterTableAddTruncateTrigger(@NAMESPACE@.slon_quote_brute(tab_nspname) || '.' || @NAMESPACE@.slon_quote_brute(tab_relname), tab_id)
				from @NAMESPACE@.sl_table
				where 2 <> (select count(*) from pg_catalog.pg_trigger,
					  pg_catalog.pg_class, pg_catalog.pg_namespace where
					  pg_trigger.tgrelid=pg_class.oid
					  AND pg_class.relnamespace=pg_namespace.oid
					  AND
					  pg_namespace.nspname = tab_nspname and tgname in ('_@CLUSTERNAME@_truncatedeny', '_@CLUSTERNAME@_truncatetrigger') and
					    pg_class.relname = tab_relname
					  );

		exception when unique_violation then
				  raise warning 'upgradeSchemaAddTruncateTriggers() - uniqueness violation';
				  raise warning 'likely due to truncate triggers existing partially';
				  raise exception 'upgradeSchemaAddTruncateTriggers() - failure - [%][%]', SQLSTATE, SQLERRM;
		end;

		-- Activate truncate triggers for replica
		perform @NAMESPACE@.alterTableConfigureTruncateTrigger(@NAMESPACE@.slon_quote_brute(tab_nspname) || '.' || @NAMESPACE@.slon_quote_brute(tab_relname)
		,'disable','enable')
			from @NAMESPACE@.sl_table
				where tab_set not in (select set_id from @NAMESPACE@.sl_set where set_origin = @NAMESPACE@.getLocalNodeId('_@CLUSTERNAME@'))
						and exists (select 1 from  pg_catalog.pg_trigger
									where pg_trigger.tgname like '_@CLUSTERNAME@_truncatetrigger' and pg_trigger.tgenabled = 'O'
									and pg_trigger.tgrelid=tab_reloid )
						and exists (select 1 from  pg_catalog.pg_trigger
									where pg_trigger.tgname like '_@CLUSTERNAME@_truncatedeny' and pg_trigger.tgenabled = 'D'
									and pg_trigger.tgrelid=tab_reloid);

		-- Activate truncate triggers for origin
		perform @NAMESPACE@.alterTableConfigureTruncateTrigger(@NAMESPACE@.slon_quote_brute(tab_nspname) || '.' || @NAMESPACE@.slon_quote_brute(tab_relname)
		,'enable','disable')
			from @NAMESPACE@.sl_table
				where tab_set in (select set_id from @NAMESPACE@.sl_set where set_origin = @NAMESPACE@.getLocalNodeId('_@CLUSTERNAME@'))
					and exists (select 1 from  pg_catalog.pg_trigger
								where pg_trigger.tgname like '_@CLUSTERNAME@_truncatetrigger' and pg_trigger.tgenabled = 'D'
								and pg_trigger.tgrelid=tab_reloid )
					and exists (select 1 from  pg_catalog.pg_trigger
								where pg_trigger.tgname like '_@CLUSTERNAME@_truncatedeny' and pg_trigger.tgenabled = 'O'
								and pg_trigger.tgrelid=tab_reloid);

		return 1;
end
$$ language plpgsql;

comment on function @NAMESPACE@.upgradeSchemaAddTruncateTriggers () is
'Add ON TRUNCATE triggers to replicated tables.';


-- ----------------------------------------------------------------------
-- FUNCTION lockSet (set_id)
--
--	Add a special trigger to all tables of a set that disables
--	access to it.
-- ----------------------------------------------------------------------
create or replace function @NAMESPACE@.lockSet (p_set_id int4)
returns int4
as $$
declare
	v_local_node_id		int4;
	v_set_row			record;
	v_tab_row			record;
begin
	-- ----
	-- Check that the set exists and that we are the origin
	-- and that it is not already locked.
	-- ----
	v_local_node_id := @NAMESPACE@.getLocalNodeId('_@CLUSTERNAME@');
	select * into v_set_row from @NAMESPACE@.sl_set
		where set_id = p_set_id
		for update;
	if not found then
		raise exception 'Slony-I: set % not found', p_set_id;
	end if;
	if v_set_row.set_origin <> v_local_node_id then
		raise exception 'Slony-I: set % does not originate on local node',
			p_set_id;
	end if;
	if v_set_row.set_locked notnull then
		raise exception 'Slony-I: set % is already locked', p_set_id;
	end if;

	-- ----
	-- Place the lockedSet trigger on all tables in the set.
	-- ----
	-- ----
	-- RBT: added 'chmac'
	-- ----
	for v_tab_row in select T.tab_id,
							@NAMESPACE@.slon_quote_brute(PGN.nspname) || '.' ||
							@NAMESPACE@.slon_quote_brute(PGC.relname) as tab_fqname,
							case when PGC.relhasmacs then ' or chmac' else '' end as onchmac
							from @NAMESPACE@.sl_table T,
								"pg_catalog".pg_class PGC, "pg_catalog".pg_namespace PGN
							where T.tab_set = p_set_id
								and T.tab_reloid = PGC.oid
								and PGC.relnamespace = PGN.oid
							order by tab_id
		loop
			execute 'create trigger "_@CLUSTERNAME@_lockedset" ' ||
					'before insert or update or delete' ||
					v_tab_row.onchmac || ' on ' ||
					v_tab_row.tab_fqname || ' for each row execute procedure
					@NAMESPACE@.lockedSet (''_@CLUSTERNAME@'');';
		end loop;

	-- ----
	-- Remember our snapshots xmax as for the set locking
	-- ----
	update @NAMESPACE@.sl_set
			set set_locked = "pg_catalog".txid_snapshot_xmax("pg_catalog".txid_current_snapshot())
			where set_id = p_set_id;

	return p_set_id;
end;
$$ language plpgsql;
comment on function @NAMESPACE@.lockSet(p_set_id int4) is
'lockSet(set_id)

Add a special trigger to all tables of a set that disables access to
it.';

-- ----------------------------------------------------------------------
-- FUNCTION alterTableAddTriggers (tab_id)
-- ----------------------------------------------------------------------
create or replace function @NAMESPACE@.alterTableAddTriggers (p_tab_id int4)
returns int4
as $$
declare
	v_no_id				int4;
	v_tab_row			record;
	v_tab_fqname		text;
	v_tab_attkind		text;
	v_n					int4;
	v_trec	record;
	v_tgbad	boolean;
begin
	-- ----
	-- Grab the central configuration lock
	-- ----
	lock table @NAMESPACE@.sl_config_lock;

	-- ----
	-- Get our local node ID
	-- ----
	v_no_id := @NAMESPACE@.getLocalNodeId('_@CLUSTERNAME@');

	-- ----
	-- Get the sl_table row and the current origin of the table.
	-- ----
	-- ----
	-- RBT: added 'chmac'
	-- ----
	select T.tab_reloid, T.tab_set, T.tab_idxname,
			S.set_origin, PGX.indexrelid,
			@NAMESPACE@.slon_quote_brute(PGN.nspname) || '.' ||
			@NAMESPACE@.slon_quote_brute(PGC.relname) as tab_fqname,
			case when PGC.relhasmacs then ' or chmac' else '' end as onchmac
			into v_tab_row
		from @NAMESPACE@.sl_table T, @NAMESPACE@.sl_set S,
			"pg_catalog".pg_class PGC, "pg_catalog".pg_namespace PGN,
			"pg_catalog".pg_index PGX, "pg_catalog".pg_class PGXC
		where T.tab_id = p_tab_id
			and T.tab_set = S.set_id
			and T.tab_reloid = PGC.oid
			and PGC.relnamespace = PGN.oid
			and PGX.indrelid = T.tab_reloid
			and PGX.indexrelid = PGXC.oid
			and PGXC.relname = T.tab_idxname
		for update;
	if not found then
		raise exception 'Slony-I: alterTableAddTriggers(): Table with id % not found', p_tab_id;
	end if;
	v_tab_fqname = v_tab_row.tab_fqname;

	v_tab_attkind := @NAMESPACE@.determineAttKindUnique(v_tab_row.tab_fqname,
														v_tab_row.tab_idxname);

	execute 'lock table ' || v_tab_fqname || ' in access exclusive mode';

	-- ----
	-- Create the log and the deny access triggers
	-- ----
	execute 'create trigger "_@CLUSTERNAME@_logtrigger"' ||
			' after insert or update or delete' ||
			v_tab_row.onchmac || ' on' ||
			v_tab_fqname || ' for each row execute procedure @NAMESPACE@.logTrigger (' ||
						pg_catalog.quote_literal('_@CLUSTERNAME@') || ',' ||
						pg_catalog.quote_literal(p_tab_id::text) || ',' ||
						pg_catalog.quote_literal(v_tab_attkind) || ');';

	execute 'create trigger "_@CLUSTERNAME@_denyaccess" ' ||
			'before insert or update or delete' ||
			v_tab_row.onchmac || ' on' ||
			v_tab_fqname || ' for each row execute procedure ' ||
			'@NAMESPACE@.denyAccess (' || pg_catalog.quote_literal('_@CLUSTERNAME@') || ');';

	perform @NAMESPACE@.alterTableAddTruncateTrigger(v_tab_fqname, p_tab_id);

	perform @NAMESPACE@.alterTableConfigureTriggers (p_tab_id);
	return p_tab_id;
end;
$$ language plpgsql;
comment on function @NAMESPACE@.alterTableAddTriggers(p_tab_id int4) is
'alterTableAddTriggers(tab_id)

Adds the log and deny access triggers to a replicated table.';

create or replace function @NAMESPACE@.recreate_log_trigger(p_fq_table_name text,
															p_tab_id oid,
															p_tab_attkind text) returns integer as $$
declare
	v_tab_row			record;
begin
	select @NAMESPACE@.slon_quote_brute(PGN.nspname) || '.' ||
		   @NAMESPACE@.slon_quote_brute(PGC.relname) as tab_fqname,
		   case when PGC.relhasmacs then ' or chmac' else '' end as onchmac
		   into v_tab_row
		   from "pg_catalog".pg_class PGC, "pg_catalog".pg_namespace PGN
		   where T.tab_id = p_tab_id;
	if not found then
		raise exception 'Slony-I: alterTableAddTriggers(): Table with id % not found', p_tab_id;
	end if;

	execute 'drop trigger "_@CLUSTERNAME@_logtrigger" on ' ||
		p_fq_table_name	;
	-- ----
	-- ----
	-- RBT: added 'chmac'
	-- ----
	execute 'create trigger "_@CLUSTERNAME@_logtrigger"' ||
			' after insert or update or delete' ||
			v_tab_row.onchmac || ' on' ||
			p_fq_table_name
			|| ' for each row execute procedure @NAMESPACE@.logTrigger (' ||
				pg_catalog.quote_literal('_@CLUSTERNAME@') || ',' ||
				pg_catalog.quote_literal(p_tab_id::text) || ',' ||
				pg_catalog.quote_literal(p_tab_attkind) || ');';
	return 0;
end
$$ language plpgsql;

comment on function  @NAMESPACE@.recreate_log_trigger(p_fq_table_name text,
	p_tab_id oid, p_tab_attkind text) is
'A function that drops and recreates the log trigger on the specified table.
It is intended to be used after the primary_key/unique index has changed.';
